/*
 * rtimetz.c: Program to determine the correct timezone for this TiVo
 * given its timezone, to calculate the offset for the timezone from
 * UTC, to obtain the UTC time from a time server, and to print out
 * a localtime strings which is suitable for use by settime.
 *
 * Copyright (c) 2004 Warren Toomey and Andrew Tridgell.
 * The program is placed under the GPL.
 */

#define _GNU_SOURCE
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>


/* External files and programs */
const char *PostcodeZones = "/etc/postcodezones";
const char *Get_Postalcode = "/hack/bin/get_postalcode";
const char *DaylightSavingsMarker = "/var/tmp/dst_lastmodified";
const char *Version= "$Revision: 1.9 $";

#define LIST_SIZE 100
char *Postcode[LIST_SIZE];
char *Timezone[LIST_SIZE];
int Listsize;
int verbose=0;


void read_postcode_zones(void)
{
  FILE *zonefile;
#define BUF_SIZE 200
  char buf[BUF_SIZE];
  char *tab;

  /* Open the file */
  zonefile = fopen(PostcodeZones, "r");
  if (zonefile == NULL) {
    if (verbose)
      fprintf(stderr, "Can't open %s: %s\n", PostcodeZones,  strerror(errno));
    else
      syslog(LOG_INFO, "Can't open %s: %s\n", PostcodeZones,  strerror(errno));
    exit(1);
  }
  /* Get each postcode and timezone */
  for (Listsize = 0; Listsize < LIST_SIZE;) {
    if (fgets(buf, BUF_SIZE - 1, zonefile) == NULL) break;

    /* Ignore comment lines */
    if (buf[0] == '#') continue;

    /* Trash the newline character */
    buf[strlen(buf) - 1] = '\0';

    /* Find the single tab in the line. Skip line if none. */
    tab = strchr(buf, '\t');
    if (tab == NULL) continue;

    /* Split the string at this point */
    *tab = '\0'; tab++;

    /* Save copies of the strings */
    Postcode[Listsize] = strdup(buf);
    Timezone[Listsize] = strdup(tab);
    Listsize++;
  }
  fclose(zonefile);
}

char *get_postcode(void)
{
  FILE *postcodepipe;
  char postcode[6] = "     ";
  char junk[100];
  int count, err;

  /* Obtain our postcode. We loop a few times in case */
  /* the MFS is in use already. */
  for (count=0; count<4; count++) {
    postcodepipe = popen(Get_Postalcode, "r");
    if (postcodepipe == NULL) {
      if (verbose)
        fprintf(stderr, "Failed to open get_postalcode pipe: %s\n",
							strerror(errno));
      else
        syslog(LOG_INFO, "Failed to open get_postalcode pipe: %s\n",
							strerror(errno));
      exit(1);
    }

    err = fread(postcode, 1, 5, postcodepipe);
    if (err != 5) {
      if (verbose)
        fprintf(stderr, "Failed to read from get_postalcode pipe: %s\n",
							strerror(errno));
        else
        syslog(LOG_INFO, "Failed to read from get_postalcode pipe: %s\n",
							strerror(errno));
      exit(1);
    }

    /* Read the rest of the pipe and close the input pipe */
    while (!feof(postcodepipe)) fread(junk, 1, 90, postcodepipe);
    fclose(postcodepipe);

    /* See if we have to retry the operation */
    if (!strcmp(postcode, "retry")) {
      sleep(2); continue;
    }

    return (strdup(postcode));
  }
  if (verbose)
    fprintf(stderr, "Failed %d times to get_postalcode\n", count);
  else
    syslog(LOG_INFO, "Failed %d times to get_postalcode\n", count);
  exit(1);
}

void set_tz_variable(char *pcode, int hoursago)
{
  int i, err;

  if (pcode == NULL) {
    if (verbose)
      fprintf(stderr, "Null postcode in set_tz_variable()\n");
    else
      syslog(LOG_INFO, "Null postcode in set_tz_variable()\n");
    exit(1);
  }

  /* Walk the list to find a matching postcode */
  for (i = 0; i < Listsize; i++) {
    if (!strcmp(Postcode[i], pcode)) {
      if (verbose && (hoursago==0))
	fprintf(stderr, "Found matching timezone %s\n", Timezone[i]);
      err = setenv("TZ", Timezone[i], 1);
      if (err == -1) {
	if (verbose)
	  fprintf(stderr, "Error setting the TZ variable: %s\n",
							strerror(errno));
 	else
	  syslog(LOG_INFO, "Error setting the TZ variable: %s\n",
							strerror(errno));
	exit(1);
      }
      return;
    }
  }
  if (verbose)
    fprintf(stderr, "No matching timezone for postcode %s\n", pcode);
  else
    syslog(LOG_INFO, "No matching timezone for postcode %s\n", pcode);
  exit(1);
}

long get_timezone_offset(char *pcode, int hoursago)
{
  time_t secs;
  struct tm *tim;
  long offset;

  /* Set the TV environment variable */
  set_tz_variable(pcode, hoursago);
  /* Get the current local time */
  secs = time(NULL);
  /* Subtract hoursago */
  secs = secs - (hoursago*3600);
  /* Localtime expects the number of GMT seconds */
  tim = localtime(&secs);
  offset = tim->tm_gmtoff;
  /* So now that we have the offset, we subtract it from secs */
  /* and get a more accurate offset */
  secs -= (time_t) offset;
  tim = localtime(&secs);
  offset = tim->tm_gmtoff;
  if (verbose && (hoursago==0))
    fprintf(stderr, "Offset %d for timezone %s, daylight saving is %d\n",
				 (int)offset, tim->tm_zone, tim->tm_isdst);
  return (offset);
}



/* Open a socket to a tcp remote host with the specified port */
static int open_socket_out(const char *host, int port)
{
  struct sockaddr_in sock_out;
  int res;
  struct in_addr addr;

  if (verbose)
    fprintf(stderr, "Connecting to time server %s\n",host);
  res = socket(PF_INET, SOCK_STREAM, 0);
  if (res == -1) {
    if (verbose)
      fprintf(stderr, "Failed to obtain a socket: %s\n", strerror(errno));
    else
      syslog(LOG_INFO, "Failed to obtain a socket: %s\n", strerror(errno));
    return(-1);
  }

  inet_aton(host, &addr);
  sock_out.sin_addr = addr;
  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *) & sock_out, sizeof(sock_out)) != 0) {
    close(res);
    if (verbose)
      fprintf(stderr, "Failed to connect to %s: %s\n", host, strerror(errno));
    else
      syslog(LOG_INFO, "Failed to connect to %s: %s\n", host, strerror(errno));
    return(-1);
  }
  return(res);
}


static void 
tivo_time(int s, long ofs)
{
  struct tm *tm;
  unsigned long t;
  if (read(s, &t, sizeof(t)) != 4) {
    if (verbose)
      fprintf(stderr, "Unable to read time from server: %s\n",strerror(errno));
    else
      syslog(LOG_INFO, "Unable to read time from server: %s\n",strerror(errno));
    exit(1);
  }

  t = ntohl(t);
  if (verbose)
    fprintf(stderr, "t was %lu from server\n", t);
  t += (ofs - 2208988800U);
  if (verbose)
    fprintf(stderr, "t is now %lu\n", t);
  tm = gmtime(&t);

  if (verbose)
    fprintf(stderr, "Time is %04d-%02d-%02d   %02d:%02d:%02d\n",
	 tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
	 tm->tm_min, tm->tm_sec);
  syslog(LOG_INFO, "Time is %04d-%02d-%02d   %02d:%02d:%02d\n",
	 tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
	 tm->tm_min, tm->tm_sec);

  printf("%04d%02d%02d%02d%02d%02d\n",
	 tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
	 tm->tm_min, tm->tm_sec);
}

void sleep_a_bit(void)
{
  int secs;
  /* Sleep for a few seconds; we don't want all the TiVos to */
  /* hit the timeserver at the same time */
  secs= time(NULL) + getpid();
  secs &= 63;
  if (verbose) {
    /* fprintf(stderr, "Pretending to sleep for for %d seconds\n", secs); */
    return;
  }
  sleep(secs);
}

void do_daylight_saving_change(long prevoffset, long offset)
{
  FILE *zout;
  struct stat sb;
  time_t secs;
  struct tm *tm;
  int err;

  if (verbose)
    fprintf(stderr, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);
  else
    syslog(LOG_INFO, "Daylight saving change: offset %ld -> %ld in secs\n",
							prevoffset,offset);

  /* We have to be careful and not adjust the clock several times a la
   * Windoze. If the DaylightSavingsMarker is more than a day old, then
   * make the change. Otherwise, don't make the change.
   */
  secs = time(NULL);
  err=stat(DaylightSavingsMarker, &sb);
  if ((err==0) && ((secs - sb.st_mtime)<86400)) {
    if (verbose)
      fprintf(stderr, "Not doing daylight saving change: I already did it\n");
    else
      syslog(LOG_INFO, "Not doing daylight saving change: I already did it\n");
    exit(0);
  }

  /* Make the change, touch the file. */
  zout= fopen(DaylightSavingsMarker, "w");
  if (zout==NULL) exit(1);
  fprintf(zout, "Daylight saving change: offset %ld -> %ld in secs\n",
                                                        prevoffset,offset);
  fclose(zout);
  
  secs= secs + (offset - prevoffset);
  tm = gmtime(&secs);
  
  if (verbose)
    fprintf(stderr, "Time is %04d-%02d-%02d   %02d:%02d:%02d\n",
         tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
         tm->tm_min, tm->tm_sec);
  syslog(LOG_INFO, "Time is %04d-%02d-%02d   %02d:%02d:%02d\n",
         tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
         tm->tm_min, tm->tm_sec);
 
  printf("%04d%02d%02d%02d%02d%02d\n",
         tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, 
         tm->tm_min, tm->tm_sec);
  exit(0);
}

int 
main(int argc, char *argv[])
{
  char *pcode;
  char *host;
  int sock;
  int servernum;
  long offset,prevoffset;
  int first_server=1;

  if (argc < 2) {
    fprintf(stderr, "rtimetz [-v] server [server ...]\n");
    exit(1);
  }

  openlog("rtimetz", LOG_PID, LOG_DAEMON);
  syslog(LOG_INFO, "rtimetz has started");

  if (!strcmp(argv[1], "-v")) {
    verbose=1; first_server=2;
  }

  read_postcode_zones();
  pcode = get_postcode();
  if (verbose)
    fprintf(stderr, "Our postcode is %s. ", pcode);
  syslog(LOG_INFO, "Our postcode is %s", pcode);
  offset = get_timezone_offset(pcode,0);
  sleep_a_bit();

  for (servernum=first_server; servernum<argc; servernum++) {
    host = argv[servernum];
    sock = open_socket_out(host, 37);
    if (sock<0) continue;
    syslog(LOG_INFO, "Connected to time server %s", host);
    tivo_time(sock, offset);
    close(sock);
    closelog();
    exit(0);
  }

  /* We failed to connect to any servers */
  if (verbose)
    fprintf(stderr, "Failed to connect to any of the time servers\n");
  else
    syslog(LOG_INFO, "Failed to connect to any of the time servers\n");

  /* Do nothing if the offset now and 2 hours ago are the same */
  prevoffset = get_timezone_offset(pcode,2);
  if (offset==prevoffset) exit(1);

  do_daylight_saving_change(prevoffset,offset);
  exit(0);
}
